/*
 * Copyright (c) 2011-2016 Pivotal Software Inc, All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package reactor.core.converter;

import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

import org.reactivestreams.Publisher;

/**
 * Base class for Reactive Streams {@link Publisher} converters.
 *
 * @author Stephane Maldini
 * @since 2.5
 */
abstract class PublisherConverter<TYPE>
		implements Function<Object, Publisher<?>>, Predicate<Object>,
		           Supplier<Class<TYPE>> {

	abstract protected Publisher<?> toPublisher(Object o);

	abstract protected TYPE fromPublisher(Publisher<?> source);

	/**
	 * Try converting a source {@link Publisher} to the target {@link Class} if supported
	 * @param source the {@link Publisher} to convert
	 * @param to the target type {@link Class}
	 *
	 * @return the converted typed or an {@link IllegalArgumentException} if failed
	 */
	public final TYPE convertTo(Publisher<?> source, Class<?> to) {
		if (get().isAssignableFrom(to)) {
			TYPE t = fromPublisher(source);

			if (t == null) {
				throw new IllegalArgumentException("Cannot convert " + source + " source to " + to + " type");
			}
			return t;
		}
		return null;
	}

	@Override
	public Publisher<?> apply(Object o) {
		Publisher<?> p = toPublisher(o);
		if (p == null) {
			throw new IllegalArgumentException("Cannot convert " + o + " source to Publisher type");
		}
		return p;
	}

	@Override
	public boolean test(Object o) {
		return get().isAssignableFrom(o.getClass());
	}
}
